"""
SAILH Canopy model.

Ported from the original SPART matlab code.

SAILH model outlined in:
    Theory of radiative transfer models applied in optical remote sensing
        - W Verhoef 1998
"""
import numpy as np
import scipy.integrate as integrate


def SAILH(soil, leafopt, canopy, angles):
    """
    Run the SAILH model.

    Parameters
    ----------
    soil : BSM.SoilOptics
        Contains soil reflectance spectra for 400 nm to 2400 nm
    leafopt : PROSPECT_5D.LeafOptics
        Contains leaf reflectance and transmittance spectra, 400 nm to 2400 nm,
        2500 to 15000 nm, and 16000 to 50000 nm.
    canopy : CanopyStructure
        Contains canopy information and SAIL model assumptions
    angles : Angles
        Holds solar zenith, observer zenith, and relative azimuth angles

    Returns
    -------
    CanopyReflectances
        Contains the four canopy reflectances arrays as attributes rso, rdo,
        rsd, rdd.
    """

    if len(leafopt.refl) != 2162:
        raise RuntimeError('Parameter leafopt.refl must be of len 2162'
                           ' i.e. include thermal specturm. \n This error'
                           ' usually occurs if you are feeding the PROSPECT_5D'
                           ' output directly into the SAILH model with adding'
                           '\n the neccessary thermal wavelengths.')

    deg2rad = np.pi / 180

    nl = canopy.nlayers
    litab = np.array([*range(5, 80, 10), *range(81, 91, 2)])[:, np.newaxis]
    LAI = canopy.LAI
    lidf = canopy.lidf
    xl = np.arange(0, -1 - (1/nl), -1 / nl)[:, np.newaxis]
    dx = 1 / nl
    iLAI = LAI * dx

    rho = leafopt.refl
    tau = leafopt.tran
    rs = soil.refl
    tts = angles.sol_angle
    tto = angles.obs_angle
    if LAI == 0:  # Just return soil reflectances and thermal set at 0.01 as
        return CanopyReflectances(rs, rs, rs, rs)

    # Set geometric quantities

    # ensures symmetry at 90 and 270 deg
    psi = abs(angles.rel_angle - 360 * round(angles.rel_angle / 360))
    psi_rad = psi * deg2rad
    sin_tts = np.sin(tts * deg2rad)
    cos_tts = np.cos(tts * deg2rad)
    tan_tts = np.tan(tts * deg2rad)

    sin_tto = np.sin(tto * deg2rad)
    cos_tto = np.cos(tto * deg2rad)
    tan_tto = np.tan(tto * deg2rad)

    sin_ttli = np.sin(litab * deg2rad)
    cos_ttli = np.cos(litab * deg2rad)

    dso = np.sqrt(tan_tts ** 2 + tan_tto ** 2 - 2 * tan_tts * tan_tto *
                  np.cos(psi_rad))

    # geometric factors associated with extinction and scattering
    chi_s, chi_o, frho, ftau = _volscatt(sin_tts, cos_tts, sin_tto, cos_tto,
                                         psi_rad, sin_ttli, cos_ttli)
    # extinction coefficient in direction of sun per
    ksli = chi_s / cos_tts  # leaf angle
    koli = chi_o / cos_tto  # observer angle
    # area scattering coefficient fractions
    sobli = frho * np.pi / (cos_tts * cos_tto)
    sofli = ftau * np.pi / (cos_tts * cos_tto)
    bfli = cos_ttli ** 2

    # integration over angles using dot product
    k = ksli.T.dot(lidf)
    K = koli.T.dot(lidf)
    bf = bfli.T.dot(lidf)
    sob = sobli.T.dot(lidf)
    sof = sofli.T.dot(lidf)

    # geometric factors for use with rho and tau
    sdb = 0.5 * (k + bf)  # specular to diffuse backward scattering
    sdf = 0.5 * (k - bf)  # specular to diffuse forward scattering
    ddb = 0.5 * (1 + bf)  # diffuse to diffuse backward scattering
    ddf = 0.5 * (1 - bf)  # diffuse to diffuse forward scattering
    dob = 0.5 * (K + bf)  # diffuse to directional backward scattering
    dof = 0.5 * (K - bf)  # diffuse to directional forward scattering

    # Probabilites
    Ps = np.exp(k * xl * LAI)  # of viewing a leaf in solar direction
    Po = np.exp(K * xl * LAI)  # of viewing a leaf in observation direction

    Ps[0:nl] = Ps[0:nl] * (1 - np.exp(-k * LAI * dx)) / (k * LAI * dx)
    Po[0:nl] = Po[0:nl] * (1 - np.exp(-k * LAI * dx)) / (k * LAI * dx)

    q = canopy.q
    Pso = np.zeros(Po.shape)

    def Psofunction(x, K, k, LAI, q, dso):
        # From APPENDIX IV of original matlab code
        if dso != 0:
            alpha = (dso / q) * 2 / (k + K)
            pso = np.exp((K + k) * LAI * x + np.sqrt(K * k) * LAI / alpha *
                         (1 - np.exp(x * alpha)))
        else:
            pso = np.exp((K + k) * LAI * x - np.sqrt(K * k) * LAI * x)

        return pso

    for j in range(len(xl)):
        Pso[j, :] = integrate.quad(Psofunction, xl[j] - dx, xl[j],
                                   args=(K, k, LAI, q, dso))[0] / dx

    # NOTE: there are two lines in the original script here that deal with
    # rounding errors. I have excluded them. If this becomes a problem see
    # lines 115 / 116 in SAILH.m

    # scattering coefficients for
    sigb = ddb * rho + ddf * tau  # diffuse backscatter incidence
    sigf = ddf * rho + ddb * tau  # forward incidence
    sb = sdb * rho + sdf * tau  # specular backscatter incidence
    sf = sdf * rho + sdb * tau  # specular forward incidence
    vb = dob * rho + dof * tau  # directional backscatter diffuse
    vf = dof * rho + dob * tau  # directional forward scatter diffuse
    w = sob * rho + sof * tau  # bidirectional scattering
    a = 1 - sigf  # attenuation
    m = np.sqrt(a ** 2 - sigb ** 2)
    rinf = (a - m) / sigb
    rinf2 = rinf * rinf

    def calcJ1(x, m, k, LAI):
        # For getting numerically stable solutions
        J1 = np.zeros((len(m), 1))
        sing = np.abs((m - k) * LAI) < 1e-6

        CS = np.where(sing)
        CN = np.where(~sing)

        J1[CN] = (np.exp(m[CN] * LAI * x) - np.exp(k * LAI * x)) / (k - m[CN])
        J1[CS] = -0.5 * (np.exp(m[CS] * LAI * x) + np.exp(k * LAI * x)) * LAI \
            * x * (1 - 1/12 * (k - m[CS]) ** 2 * LAI ** 2)
        return J1

    def calcJ2(x, m, k, LAI):
        # For getting numerically stable solutions
        J2 = (np.exp(k * LAI * x) - np.exp(-k * LAI) * np.exp(-m * LAI *
                                                              (1 + x))
              ) / (k + m)
        return J2

    # direct solar radiation
    J1k = calcJ1(-1, m, k, LAI)
    J2k = calcJ2(0, m, k, LAI)
    J1K = calcJ1(-1, m, K, LAI)
    J2K = calcJ2(0, m, K, LAI)

    e1 = np.exp(-m * LAI)
    e2 = e1 ** 2
    re = rinf * e1

    denom = 1 - rinf2 ** 2

    s1 = sf + rinf * sb
    s2 = sf * rinf + sb
    v1 = vf + rinf * vb
    v2 = vf * rinf + vb
    Pss = s1 * J1k
    Qss = s2 * J2k
    Poo = v1 * J1K
    Qoo = v2 * J2K

    tau_ss = np.exp(-k * LAI)
    tau_oo = np.exp(-K * LAI)

    Z = (1 - tau_ss * tau_oo) / (K + k)

    tau_dd = (1 - rinf2) * e1 / denom
    rho_dd = rinf * (1 - e2) / denom
    tau_sd = (Pss - re * Qss) / denom
    tau_do = (Poo - re * Qoo) / denom
    rho_sd = (Qss - re * Pss) / denom
    rho_do = (Qoo - re * Poo) / denom

    T1 = v2 * s1 * (Z - J1k * tau_oo) / (K + m) + v1 * s2 * (Z - J1K * tau_ss)\
        / (k + m)
    T2 = -(Qoo * rho_sd + Poo * tau_sd) * rinf
    rho_sod = (T1 + T2) / (1 - rinf2)

    rho_sos = w * np.sum(Pso[0: nl]) * iLAI
    rho_so = rho_sod + rho_sos

    Pso2w = Pso[nl]

    # Sail analytical reflectances
    denom = 1 - rs * rho_dd

    rso = rho_so + rs * Pso2w + (
        (tau_sd + tau_ss * rs * rho_dd) * tau_oo + (tau_sd + tau_ss) * tau_do)\
        * rs / denom
    rdo = rho_do + (tau_oo + tau_do) * rs * tau_dd / denom
    rsd = rho_sd + (tau_ss + tau_sd) * rs * tau_dd / denom
    rdd = rho_dd + tau_dd * rs * tau_dd / denom

    rad = CanopyReflectances(rso, rdo, rsd, rdd)

    return rad


class CanopyReflectances:
    """CanopyReflectances"""
    """Class to hold canopy reflectances computed by the SAILH model.

    Parameters
    ----------
    rso : np.array
        Bidirectional reflectance of the canopy
    rdo : np.array
        Directional reflectance for diffuse incidence of the canopy
    rsd : np.array
        Diffuse reflectance for specular incidence of the canopy
    rdd : np.array
        Diffuse reflectance for diffuse incidence of the canopy

    Attributes
    ----------
    rso : np.array
        Bidirectional reflectance of the canopy
    rdo : np.array
        Directional reflectance for diffuse incidence of the canopy
    rsd : np.array
        Diffuse reflectance for specular incidence of the canopy
    rdd : np.array
        Diffuse reflectance for diffuse incidence of the canopy
    """
    def __init__(self, rso, rdo, rsd, rdd):
        self.rso = rso
        self.rdo = rdo
        self.rsd = rsd
        self.rdd = rdd


class Angles:
    """Class to hold solar zenith, observation zenith, and relative azimuth
    angles.

    Parameters
    ----------
    sol_angle : float
        Solar zenith angle, degrees
    obs_angle : float
        Observer zenith angle, degrees
    rel_angle : float
        Relative azimuth angle, degrees

    Attributes
    ----------
    sol_angle : float
        Solar zenith angle, degrees
    obs_angle : float
        Observer zenith angle, degrees
    rel_angle : float
        Relative azimuth angle, degrees
    """
    def __init__(self, sol_angle, obs_angle, rel_angle):
        self.sol_angle = sol_angle
        self.obs_angle = obs_angle
        self.rel_angle = rel_angle


class CanopyStructure:
    """
    Class to hold canopy properties. Some are user specified, others are
    SAIL model assumptions.

    Parameters
    ----------
    LAI : float
        Leaf area index, 0 to 8
    LIDFa : float
        Leaf inclination distribution function parameter a, range -1 to 1
    LIDFb : float
        Leaf inclination distribution function parameter b, range -1 to 1
    q : float
        Canopy hotspot parameter: leaf width / canopy height, range 0 to 0.2

    Attributes
    ----------
    LAI : float
        Leaf area index, 0 to 8
    LIDFa : float
        Leaf inclination distribution function parameter a, range -1 to 1
    LIDFb : float
        Leaf inclination distribution function parameter b, range -1 to 1
    q : float
        Canopy hotspot parameter: leaf width / canopy height, range 0 to 0.2
    nlayers : int
        Number of layers in canopy, 60 (SAIL assumption)
    nlincl : int
        Number of different leaf inclination angles, 13 (SAIL assumption)
    nlazi : int
        Number of different leaf azimuth angles, 36 (SAIL assumption)
    lidf : np.array
        Leaf inclination distribution function, calculated from LIDF params
    """
    def __init__(self, LAI, LIDFa, LIDFb, q):
        self.LAI = LAI
        self.LIDFa = LIDFa
        self.LIDFb = LIDFb
        self.q = q
        self.nlayers = 60
        self.nlincl = 13
        self.nlazi = 36
        self.lidf = calculate_leafangles(LIDFa, LIDFb)


def calculate_leafangles(LIDFa, LIDFb):
    """Calculate the Leaf Inclination Distribution Function as outlined
    by Verhoef in paper cited at the top of this script.

    Parameters
    ----------
    LIDFa : float
        Leaf inclination distribution function parameter a, range -1 to 1
    LIDFb : float
        Leaf inclination distribution function parameter b, range -1 to 1

    Returns
    -------
    np.array
        Leaf inclination distribution function, calculated from LIDF
    """
    def dcum(a, b, theta):
        # Calculate cumulative distribution
        rd = np.pi / 180
        if LIDFa > 1:
            f = 1 - np.cos(theta * rd)
        else:
            eps = 1e-8
            delx = 1
            x = 2 * rd * theta
            theta2 = x
            while delx > eps:
                y = a * np.sin(x) + 0.5 * b * np.sin(2 * x)
                dx = 0.5 * (y - x + theta2)
                x = x + dx
                delx = abs(dx)
            f = (2 * y + theta2) / np.pi
        return f

    # F sized to 14 entries so diff for actual LIDF becomes 13 entries
    F = np.zeros((14, 1))
    for i in range(1, 9):
        theta = i * 10
        F[i] = dcum(LIDFa, LIDFb, theta)
    for i in range(9, 13):
        theta = 80 + (i - 8) * 2
        F[i] = dcum(LIDFa, LIDFb, theta)
    F[13] = 1

    lidf = np.diff(F, axis=0)

    return lidf


def _volscatt(sin_tts, cos_tts, sin_tto, cos_tto, psi_rad, sin_ttli, cos_ttli):
    # Calculate geometric factors. See SAILH.m code.
    # See original matlab code. Adapted here to save recalculating trigs.
    nli = len(cos_ttli)

    psi_rad = psi_rad * np.ones((nli, 1))
    cos_psi = np.cos(psi_rad)

    Cs = cos_ttli * cos_tts
    Ss = sin_ttli * sin_tts

    Co = cos_ttli * cos_tto
    So = sin_ttli * sin_tto

    As = np.maximum(Ss, Cs)
    Ao = np.maximum(So, Co)

    bts = np.arccos(-Cs / As)
    bto = np.arccos(-Co / Ao)

    chi_o = 2 / np.pi * ((bto - np.pi / 2) * Co + np.sin(bto) * So)
    chi_s = 2 / np.pi * ((bts - np.pi / 2) * Cs + np.sin(bts) * Ss)

    delta1 = np.abs(bts - bto)
    delta2 = np.pi - np.abs(bts + bto - np.pi)

    Tot = psi_rad + delta1 + delta2

    bt1 = np.minimum(psi_rad, delta1)
    bt3 = np.maximum(psi_rad, delta2)
    bt2 = Tot - bt1 - bt3

    T1 = 2 * Cs * Co + Ss * So * cos_psi
    T2 = np.sin(bt2) * (2 * As * Ao + Ss * So * np.cos(bt1) * np.cos(bt3))

    Jmin = bt2 * T1 - T2
    Jplus = (np.pi - bt2) * T1 + T2

    frho = Jplus / (2 * np.pi ** 2)
    ftau = -Jmin / (2 * np.pi ** 2)

    zeros = np.zeros((nli, 1))
    frho = np.maximum(zeros, frho)
    ftau = np.maximum(zeros, ftau)

    return chi_s, chi_o, frho, ftau



"""
SMAC model.

Library for atmospheric correction using SMAC method (Rahman and Dedieu, 1994)
Original translated and improved by:
Peiqi Yang, ITC, University of Twente

Adapted to Python by:
George Worrall, CRS, University of Florida
"""
import numpy as np


def SMAC(angles, atm, coefs):
    """
    Run the SMAC atmosphere model.

    Parameters
    ----------
    angles : SAILH.Angles
        Contains solar zenith, observational zentiy, and relative difference of
        azimuth angles in degrees.
    atm : AtmosphericProperties
        Contains the atmospheric properties for the model.
    coefs : dict
        Contains all the transmittance coefficients for different atmospheric
        gases at the wavelengths of the specified sensor.

    Returns
    -------
    AtmosphericOptics
        Contains atmospheric reflectance and transmittance arrays
    """
    # Extract variables for more concise use below.
    # Model coefficients for days.
    tts = angles.sol_angle
    tto = angles.obs_angle
    psi = angles.rel_angle
    Pa = atm.Pa
    taup550 = atm.aot550
    uo3 = atm.uo3
    uh2o = atm.uh2o

    ah2o = coefs['ah2o']
    nh2o = coefs['nh2o']
    ao3 = coefs['ao3']
    no3 = coefs['no3']
    ao2 = coefs['ao2']
    no2 = coefs['no2']
    po2 = coefs['po2']
    aco2 = coefs['aco2']
    nco2 = coefs['nco2']
    pco2 = coefs['pco2']
    ach4 = coefs['ach4']
    nch4 = coefs['nch4']
    pch4 = coefs['pch4']
    ano2 = coefs['ano2']
    nno2 = coefs['nno2']
    pno2 = coefs['pno2']
    aco = coefs['aco']
    nco = coefs['nco']
    pco = coefs['pco']
    a0s = coefs['a0s']
    a1s = coefs['a1s']
    a2s = coefs['a2s']
    a3s = coefs['a3s']
    a0T = coefs['a0T']
    a1T = coefs['a1T']
    a2T = coefs['a2T']
    a3T = coefs['a3T']
    taur = coefs['taur']
    # sr = coefs['sr']
    a0taup = coefs['a0taup']
    a1taup = coefs['a1taup']
    wo = coefs['wo']
    gc = coefs['gc']
    a0P = coefs['a0P']
    a1P = coefs['a1P']
    a2P = coefs['a2P']
    a3P = coefs['a3P']
    a4P = coefs['a4P']
    Rest1 = coefs['Rest1']
    Rest2 = coefs['Rest2']
    Rest3 = coefs['Rest3']
    Rest4 = coefs['Rest4']
    Resr1 = coefs['Resr1']
    Resr2 = coefs['Resr2']
    Resr3 = coefs['Resr3']
    Resa1 = coefs['Resa1']
    Resa2 = coefs['Resa2']
    Resa3 = coefs['Resa3']
    Resa4 = coefs['Resa4']

    cdr = np.pi / 180
    crd = 180 / np.pi

    # Calculate TOA reflectance
    us = np.cos(tts * cdr)
    uv = np.cos(tto * cdr)
    Peq = Pa / 1013.25

    m = 1 / us + 1 / uv  # air mass
    taup = a0taup + a1taup * taup550  # aerosol optical depth in spectral band

    uo2 = Peq ** po2  # gaseous transmissions (down and up paths)
    uco2 = Peq ** pco2
    uch4 = Peq ** pch4
    uno2 = Peq ** pno2
    uco = Peq ** pco

    to3 = np.exp(ao3 * (uo3 * m) ** no3)
    th2o = np.exp(ah2o * (uh2o * m) ** nh2o)
    to2 = np.exp(ao2 * (uo2 * m) ** no2)
    tco2 = np.exp(aco2 * (uco2 * m) ** nco2)
    tch4 = np.exp(ach4 * (uch4 * m) ** nch4)
    tno2 = np.exp(ano2 * (uno2 * m) ** nno2)
    tco = np.exp(aco * (uco * m) ** nco)

    tg = th2o * to3 * to2 * tco2 * tch4 * tco * tno2  # Eq 6

    # spherical albedo of the atmosphere
    s = a0s * Peq + a3s + a1s * taup550 + a2s * taup550 ** 2  # mod of Eq 8

    # total scattering transmission
    ttetas = a0T + a1T * taup550 / us + (a2T * Peq + a3T) / (1 + us)
    ttetav = a0T + a1T * taup550 / uv + (a2T * Peq + a3T) / (1 + uv)

    # scattering angle cosine
    cksi = - ((us * uv) + (np.sqrt(1 - us * us) * np.sqrt(1 - uv * uv) *
                           np.cos(psi * crd)))

    # hard limit on cksi -> from original matlab
    if cksi < -1:
        cksi = -1

    # scattering angle in degrees
    ksiD = crd * np.arccos(cksi)

    # rayleigh atmospheric reflectance
    ray_phase = 0.7190443 * (1 + (cksi * cksi)) + 0.0412742  # Eq 13
    ray_ref = (taur * ray_phase) / (4 * us * uv)  # Eq 11
    ray_ref = ray_ref * Pa / 1013.25  # correction for pressure variation
    taurz = taur * Peq  # Eq 12

    aer_phase = a0P + a1P * ksiD + a2P * ksiD * ksiD + a3P * ksiD ** 3 + \
        a4P * ksiD ** 4  # extension of Eq 17
    ak2 = (1 - wo) * (3 - wo * 3 * gc)
    ak = np.sqrt(ak2)

    # X Y Z Appendix
    e = -3 * us * us * wo / (4 * (1 - ak2 * us * us))
    f = -(1 - wo) * 3 * gc * us * us * wo / (4 * (1 - ak2 * us * us))
    dp = e / (3 * us) + us * f
    d = e + f
    b = 2 * ak / (3 - wo * 3 * gc)
    delta = np.exp(ak * taup) * (1 + b) ** 2 - np.exp(-ak * taup) * \
        (1 - b) ** 2
    ww = wo / 4
    ss = us / (1 - ak2 * us * us)
    q1 = 2 + 3 * us + (1 - wo) * 3 * gc * us * (1 + 2 * us)
    q2 = 2 - 3 * us - (1 - wo) * 3 * gc * us * (1 - 2 * us)
    q3 = q2 * np.exp(-taup / us)
    c1 = ((ww * ss) / delta) * (q1 * np.exp(ak * taup) * (1 + b) + q3 *
                                (1 - b))
    c2 = -((ww * ss) / delta) * (q1 * np.exp(-ak * taup) * (1 - b) + q3 *
                                 (1 + b))
    cp1 = c1 * ak / (3 - wo * 3 * gc)
    cp2 = -c2 * ak / (3 - wo * 3 * gc)
    z = d - wo * 3 * gc * uv * dp + wo * aer_phase / 4
    x = c1 - wo * 3 * gc * uv * cp1
    y = c2 - wo * 3 * gc * uv * cp2
    aa1 = uv / (1 + ak * uv)
    aa2 = uv / (1 - ak * uv)
    aa3 = us * uv / (us + uv)

    aer_ref1 = x * aa1 * (1 - np.exp(-taup / aa1))  # Eq 13
    aer_ref2 = y * aa2 * (1 - np.exp(-taup / aa2))
    aer_ref3 = z * aa3 * (1 - np.exp(-taup / aa3))

    aer_ref = (aer_ref1 + aer_ref2 + aer_ref3) / (us * uv)

    # Residual rayleigh (not in the paper)
    Res_ray = Resr1 + Resr2 * taur * ray_phase / (us * uv) + Resr3 * \
        ((taur * ray_phase / (us * uv)) ** 2)

    # Residual aerosol
    Res_aer = (Resa1 + Resa2 * (taup * m * cksi) + Resa3 *
               ((taup * m * cksi) ** 2)) + Resa4 * (taup*m*cksi) ** 3

    # Term coupling molecule / aerosol
    tautot = taup + taurz

    Res_6s = (Rest1 + Rest2 * (tautot * m * cksi) + Rest3 *
              ((tautot * m * cksi) ** 2)) + Rest4 * ((tautot * m * cksi) ** 3)

    # Total atmospheric reflectance
    atm_ref = ray_ref - Res_ray + aer_ref - Res_aer + Res_6s

    # For non-lambertian surface
    tdir_tts = np.exp(-tautot / us)
    tdir_ttv = np.exp(-tautot / uv)
    tdif_tts = ttetas - tdir_tts
    tdif_ttv = ttetav - tdir_ttv

    atm_opts = AtmosphericOptics(ttetas, ttetav, tg, s, atm_ref, tdir_tts,
                                 tdif_tts, tdir_ttv, tdif_ttv)

    return atm_opts


class AtmosphericOptics:
    """
    Class to hold atmospheric optics results from the SMAC model.

    Parameters
    ----------
    Ta_s : np.array
        Total scattering transmission downard
    Ta_o : np.array
        Total scattering transmission upward
    Tg : np.array
        Transmittance for all gases
    Ra_dd : np.array
        Hemispherical atmospheric reflectance for diffuse light
    Ra_so : np.array
        Directional atmospheric reflectance for direct incidence
    Ta_ss : np.array
        Directional transmittance for direct incidence
    Ta_sd : np.array
        Hemispherical transmittance for direct incidence
    Ta_oo : np.array
        Directional transmittance for direct incidence (in viewing direction)
    Ta_do : np.array
        Hemispherical transmittance for direct incidence (in viewing direction)

    Attributes
    ----------
    Ta_s : np.array
        Total scattering transmission downard
    Ta_o : np.array
        Total scattering transmission upward
    Tg : np.array
        Transmittance for all gases
    Ra_dd : np.array
        Hemispherical atmospheric reflectance for diffuse light
    Ra_so : np.array
        Directional atmospheric reflectance for direct incidence
    Ta_ss : np.array
        Directional transmittance for direct incidence
    Ta_sd : np.array
        Hemispherical transmittance for direct incidence
    Ta_oo : np.array
        Directional transmittance for direct incidence (in viewing direction)
    Ta_do : np.array
        Hemispherical transmittance for direct incidence (in viewing direction)
    """
    def __init__(self, Ta_s, Ta_o, Tg, Ra_dd, Ra_so, Ta_ss, Ta_sd, Ta_oo,
                 Ta_do):
        self.Ta_s = Ta_s
        self.Ta_o = Ta_o
        self.Tg = Tg
        self.Ra_dd = Ra_dd
        self.Ra_so = Ra_so
        self.Ta_ss = Ta_ss
        self.Ta_sd = Ta_sd
        self.Ta_oo = Ta_oo
        self.Ta_do = Ta_do


class AtmosphericProperties:
    """
    Class to hold the properties of the atmosphere used in SMAC calculations.

    Parameters
    ----------
    aot550 : float
        Aerosol optical thickness at 550 nm
    uo3 : float
        Ozone content, cm-atm
    uh2o : float
        Water vapour content, g cm^-2
    Pa : float, optional
        Air pressure, hPa, defaults to 1013.25
    alt_m : int, optional
        Altitude of observation site, used to calculate air pressure if air
        pressure not known.
    temp_k : float, optional
        Temperature in kelvin used to estimate air pressure

    Attributes
    ----------
    aot550 : float
        Aerosol optical thickness at 550 nm
    uO3 : float
        Ozone content, cm-atm
    uh20 : float
        Water vapour content, g cm^-2
    Pa : float
        Air pressure, hPa
    """
    def __init__(self, aot550, uo3, uh2o, Pa=None, alt_m=None, temp_k=None):
        self.aot550 = aot550
        self.uo3 = uo3
        self.uh2o = uh2o
        if isinstance(Pa, type(None)):
            if not isinstance(alt_m, type(None)) and not isinstance(
                    temp_k, type(None)):
                self.Pa = _calculate_pressure_from_altitude(alt_m, temp_k)
            else:
                self.Pa = 1013.25
        else:
            self.Pa = Pa


def _calculate_pressure_from_altitude(alt_m, temp_k):
    # Caluclate pressure from altitude and temperature

    g = 9.80665
    M = 0.02896968
    R0 = 8.314462618
    Pa0 = 1013.25

    Pa = Pa0 * np.exp(-(g * alt_m * M / (temp_k * R0)))

    return Pa



"""
SPART-python

PROSPECT 5D model.

Feret et al. - PROSPECT-D: Towards modeling leaf optical properties
    through a complete lifecycle
"""
import numpy as np
import scipy.integrate as integrate


class LeafBiology:
    """
    Class to hold leaf biology variables.

    Parameters
    ----------
    Cab : float
        Chlorophyll concentration, micro g / cm ^ 2
    Cca : float
        Carotenoid concentration, micro g / cm ^ 2
    Cw : float
        Equivalent water thickness, cm
    Cdm : float
        Leaf mass per unit area, g / cm ^ 2
    Cs : float
        Brown pigments (from SPART paper, unitless)
    Cant : float
        Anthocyanin content, micro g / cm ^ 2
    N : float
        Leaf structure parameter. Unitless.

    Attributes
    ----------
    Cab : float
        Chlorophyll concentration, micro g / cm ^ 2
    Cca : float
        Carotenoid concentration, micro g / cm ^ 2
    Cw : float
        Equivalent water thickness, cm
    Cdm : float
        Leaf mass per unit area, g / cm ^ 2
    Cs : float
        Fraction senescent material (brown pigment from SPART paper, unitless)
    Cant : float
        Anthocyanin content, micro g / cm ^ 2
    N : float
        Leaf structure parameter. Unitless.
    rho_thermal : float
        Reflectance in the thermal range. SPART assumption: 0.01
    tau_thermal : float
        Transmittance in the thermal range. SPART assumption: 0.01
    """

    def __init__(self, Cab, Cca, Cw, Cdm, Cs, Cant, N):
        self.Cab = Cab
        self.Cca = Cca
        self.Cw = Cw
        self.Cdm = Cdm
        self.Cs = Cs
        self.Cant = Cant
        self.N = N
        self.rho_thermal = 0.01
        self.tau_thermal = 0.01


class LeafOptics:
    """
    Class to hold leaf optics information.

    Parameters
    ----------
    refl : np.array
        Spectral reflectance of the leaf, 400 to 2400 nm
    tran : np.array
        Spectral transmittance of the leaf, 400 to 2400 nm
    kChlrel : np.array
        Relative portion of chlorophyll contribution to reflecntace
        / transmittance in the spectral range, 400 to 2400 nm

    Attributes
    ----------
    refl : np.array
        Spectral reflectance of the leaf, 400 to 2400 nm
    tran : np.array
        Spectral transmittance of the leaf, 400 to 2400 nm
    kChlrel : np.array
        Relative portion of chlorophyll contribution to reflecntace
        / transmittance in the spectral range, 400 to 2400 nm
    """
    def __init__(self, refl, tran, kChlrel):
        self.refl = refl
        self.tran = tran
        self.kChlrel = kChlrel


def PROSPECT_5D(leafbio, optical_params):
    """
    PROSPECT_5D model.

    Parameters
    ----------
    leafbio : LeafBiology
        Object holding user specified leaf biology model parameters.
    optical_params : dict
        Optical parameter constants. Loaded externally and passed in.

    Returns
    -------
    LeafOptics
        Contains attributes relf, tran, kChlrel for reflectance, transmittance
        and contribution of chlorophyll over the 400 nm to 2400 nm spectrum
    """
    # Leaf parameters
    Cab = leafbio.Cab
    Cca = leafbio.Cca
    Cw = leafbio.Cw
    Cdm = leafbio.Cdm
    Cs = leafbio.Cs
    Cant = leafbio.Cant
    N = leafbio.N

    # Model constants
    nr = optical_params['nr']
    Kdm = optical_params['Kdm']
    Kab = optical_params['Kab']
    Kca = optical_params['Kca']
    Kw = optical_params['Kw']
    Ks = optical_params['Ks']
    Kant = optical_params['Kant']

    # Compact leaf layer
    Kall = (Cab * Kab + Cca * Kca + Cdm * Kdm + Cw * Kw + Cs * Ks +
            Cant * Kant) / N

    # Non-conservative scattering (normal case)
    j = np.where(Kall > 0)[0]
    t1 = (1 - Kall) * np.exp(-Kall)

    def expint(x):
        # NOTE: differences in final output come from this integral
        # which evaluates slightly different (10 decimal places) than matlab
        # Exponential integral from expint command in matlab
        def intergrand(t):
            return np.exp(-t) / t
        return integrate.quad(intergrand, x, np.inf)

    t2 = Kall ** 2 * np.vectorize(expint)(Kall)[0]
    tau = np.ones((len(t1), 1))
    tau[j] = t1[j] + t2[j]
    kChlrel = np.zeros((len(t1), 1))
    kChlrel[j] = Cab * Kab[j] / (Kall[j] * N)

    t_alph = calculate_tav(40, nr)
    r_alph = 1 - t_alph
    t12 = calculate_tav(90, nr)
    r12 = 1 - t12
    t21 = t12 / (nr ** 2)
    r21 = 1 - t21

    # top surface side
    denom = 1 - r21 * r21 * tau ** 2
    Ta = t_alph * tau * t21 / denom
    Ra = r_alph + r21 * tau * Ta

    # bottom surface side
    t = t12 * tau * t21 / denom
    r = r12 + r21 * tau * t

    # Stokes equations to compute properties of next N-1 layers (N real)

    # Normal case
    D = np.sqrt((1 + r + t) * (1 + r - t) * (1 - r + t) * (1 - r - t))
    rq = r ** 2
    tq = t ** 2
    a = (1 + rq - tq + D) / (2 * r)
    b = (1 - rq + tq + D) / (2 * t)

    bNm1 = b ** (N - 1)
    bN2 = bNm1 ** 2
    a2 = a ** 2
    denom = a2 * bN2 - 1
    Rsub = a * (bN2 - 1) / denom
    Tsub = bNm1 * (a2 - 1) / denom

    # Case of zero absorption
    j = np.where(r + t >= 1)[0]
    Tsub[j] = t[j] / (t[j] + (1 - t[j]) * (N - 1))
    Rsub[j] = 1 - Tsub[j]

    # Reflectance and transmittance of the leaf:
    #   combine top llayer with next N-1 layers
    denom = 1 - Rsub * r
    tran = Ta * Tsub / denom
    refl = Ra + Ta * Rsub * t / denom

    leafopt = LeafOptics(refl, tran, kChlrel)

    return leafopt


def calculate_tav(alpha, nr):
    """
    Calculate average transmissitivity of a dieletrie plane surface.

    Parameters
    ----------
    alpha : float
        Maximum incidence angle defining the solid angle.
    nr : float
        Refractive index

    Returns
    -------
    float
        Transmissivity of a dielectric plane surface averages over all
        directions of incidence and all polarizations.

    NOTE
    ----
    Lifted directly from original SPART matlab calculations.
    Papers cited in original PROSPECT model:
    Willstatter-Stoll Theory of Leaf Reflectance Evaluated
    by Ray Tracinga - Allen et al.
    Transmission of isotropic radiation across an interface
    between two dielectrics - Stern
    """
    rd = np.pi / 180
    n2 = nr ** 2
    n_p = n2 + 1
    nm = n2 - 1
    a = (nr + 1) * (nr + 1) / 2
    k = -(n2 - 1) * (n2 - 1) / 4
    sa = np.sin(alpha * rd)

    b1 = 0
    if alpha != 90:
        b1 = np.sqrt((sa ** 2 - n_p / 2) * (sa ** 2 - n_p / 2) + k)
    b2 = sa ** 2 - n_p / 2
    b = b1 - b2
    b3 = b ** 3
    a3 = a ** 3
    ts = (k ** 2 / (6 * b3) + k / b - b / 2) - (k ** 2 / (6 * a3) + k / a
                                                - a / 2)

    tp1 = -2 * n2 * (b - a) / (n_p ** 2)
    tp2 = -2 * n2 * n_p * np.log(b / a) / (nm ** 2)
    tp3 = n2 * (1 / b - 1 / a) / 2
    tp4 = (16 * n2 ** 2 * (n2 ** 2 + 1) * np.log((2 * n_p * b - nm ** 2) /
                                                 (2 * n_p * a - nm ** 2))
           / (n_p ** 3 * nm ** 2))
    tp5 = (16 * n2 ** 3 * (1 / (2 * n_p * b - nm ** 2) - 1 / (2 * n_p * a -
                                                              nm ** 2))
           / n_p ** 3)
    tp = tp1 + tp2 + tp3 + tp4 + tp5
    tav = (ts + tp) / (2 * sa ** 2)

    return tav


if __name__ == '__main__':
    # Test cases, compare to original matlab outputs
    from SPART import load_optical_parameters
    leafbio = LeafBiology(40, 10, 0.02, 0.01, 0, 10, 1.5)
    leafopt = PROSPECT_5D(leafbio, load_optical_parameters())
    print(leafopt.tran)
    print(leafopt.refl)
    print(leafopt.kChlrel)



"""
Soil-Plant-Atmosphere Radiative Transfer model
for top-of-canopy and top-of-atmosphere reflectance

Coupling BSM, PROSAIL and SMAC to simulate TOA reflectance

Python port coded by George Worrall (gworrall@ufl.edu)
Center for Remote Sensing, University of Florida

Ported to Python from the original matlab code and model developed by:

Peiqi Yang               (p.yang@utwente.nl)
Christiaan van der Tol   (c.vandertol@utwente.nl)
Wout Verhoef             (w.verhoef@utwente.nl)

University of Twente
Faculty of Geo-Information Science and Earth Observation (ITC),
Department of Water Resources
"""
import pickle
import numpy as np
import pandas as pd
from pathlib import Path
from .BSM import BSM, SoilParameters, SoilParametersFromFile
from .PROSPECT_5D import PROSPECT_5D, LeafBiology
from .SAILH import SAILH, CanopyStructure, Angles
from .SMAC import SMAC, AtmosphericProperties

# TODO: look at threading BSM and PROSPECT as they don't require each
# other

class SPART:
    """
    SPART model.

    Parameters
    ----------
    soilpar : SoilParameters
        Holds the soil parameters for the BSM model.
    leafbio : LeafBiology
        Hold the leaf biology parameters for the PROSPECT model.
    canopy : CanopyStructure
        Holds the canopy parameters for the SAILH model.
    atm : AtmosphericProperties
        Holds the atmospheric properties for the SMAC model.
    sensor : str
        Name of RS platform to simulate. This is done after the SAIL stage and
        saves on atmospheric diffuse reflectance calculations for TOC and
        atmospheric correction calculations for TOA.
        Currently available:
        'TerraAqua-MODIS'
        'LANDSAT4-TM'
        'LANDSAT5-TM'
        'LANDSAT7-ETM'
        'LANDSAT8-OLI'
        'Sentinel3A-OLCI'
        'Sentinel3B-OLCI'
    DOY : int
        Day of the year - Julian calendar

    Attributes
    ----------
    soilopt : BSM.SoilOptics
        Contains soil reflectances
    leafopt : PROSPECT_5D.LeafOptics
        Contains leaf reflectance and transmittance and fraction contributed
        by chlorphyll
    canopyopt : SAILH.CanopyReflectances
        Contains bidrectional and directional, diffuce and specular reflectance
    R_TOC : np.array
        Top of canopy reflectance
    R_TOA : np.array
        Top of atmosphere reflectance
    L_TOA : np.array
        Top of atmosphere radiance
    """
    def __init__(self, soilpar, leafbio, canopy, atm, angles, sensor, DOY):
        self._tracker = {}  # tracks changes to input parameters
        self.soilpar = soilpar
        self.leafbio = leafbio
        self.canopy = canopy
        self.atm = atm
        self.angles = angles
        self.sensor = sensor
        self.DOY = DOY
        self.spectral = SpectralBands()
        self.optipar = load_optical_parameters()
        self.ETpar = load_ET_parameters()
        self.sensorinfo = load_sensor_info(sensor)

        # Initialize arrays

        # Leaf reflectance array, spans 400 to 2400 nm in 1 nm increments
        # then 2500 to 15000 in 100 nm increments
        # then 16000 to 50000 in 1000 nm increments
        self._rho = np.zeros((self.spectral.nwlP + self.spectral.nwlT, 1))
        # Leaf transmittance array, as above
        self._tau = np.zeros((self.spectral.nwlP + self.spectral.nwlT, 1))
        # Soil reflectance array, as above
        self._rsoil = np.zeros((self.spectral.nwlP + self.spectral.nwlT, 1))

        # Set the model reflectance and transmittance assumptions
        self.set_refl_trans_assumptions()


    @property
    def soilpar(self):
        return self._soilpar

    @soilpar.setter
    def soilpar(self, soilpar):
        self._soilpar = soilpar
        self._tracker['soil'] = True

    @property
    def leafbio(self):
        return self._leafbio

    @leafbio.setter
    def leafbio(self, leafbio):
        self._leafbio = leafbio
        self._tracker['leaf'] = True

    @property
    def canopy(self):
        return self._canopy

    @canopy.setter
    def canopy(self, canopy):
        self._canopy = canopy
        self._tracker['canp'] = True

    @property
    def atm(self):
        return self._atm

    @atm.setter
    def atm(self, atm):
        self._atm = atm
        self._tracker['atm'] = True

    @property
    def angles(self):
        return self._angles

    @angles.setter
    def angles(self, angles):
        self._angles = angles
        self._tracker['angles'] = True

    @property
    def sensor(self):
        return self._sensor

    @sensor.setter
    def sensor(self, sensor):
        self._sensor = sensor
        self._tracker['sensor'] = True

    @property
    def DOY(self):
        return self._DOY

    @DOY.setter
    def DOY(self, DOY):
        self._DOY = DOY
        self._tracker['DOY'] = True

    def set_refl_trans_assumptions(self):
        """Sets the model assumptions about soil and leaf reflectance and
        transmittance in the thermal range.

        These are that soil reflectance is the value for 2400 nm
        in the entire thermal range and that leaf relctance and
        transmittance are 0.01 in the thermal range (this is
        a model assumption that is set in the LeafBiology class in the BSM
        script)

        Returns
        -------
        None
        """
        self._rho[self.spectral.IwlT] = self.leafbio.rho_thermal
        self._tau[self.spectral.IwlT] = self.leafbio.tau_thermal
        self._rsoil[self.spectral.IwlT] = 1

    def run(self):
        """Run the SPART model.

        Returns
        -------
        pd.DataFrame
            Contains the radiances and reflectances columns 'Band' 'L_TOA'
            'R_TOA' 'R_TOC' index by central band wavelength
        """
        # Calculate ET radiance from the sun for look angles and DOY
        if self._tracker['DOY'] or self._tracker['angles']:
            # Calculate extra-terrestrial radiance for the day
            Ra = calculate_ET_radiance(self.ETpar['Ea'], self.DOY,
                                       self.angles.sol_angle)
            self._La = calculate_spectral_convolution(self.ETpar['wl_Ea'], Ra,
                                                      self.sensorinfo)
            self._tracker['DOY'] = False
            self._tracker['angles'] = False

        # Run the BSM model
        if self._tracker['soil']:
            soilopt = BSM(self._soilpar, self.optipar)
            # Update soil optics refl and trans to include thermal
            # values from model assumptions
            self._rsoil[self.spectral.IwlP] = soilopt.refl
            self._rsoil[self.spectral.IwlT] = 1 * self._rsoil[self.spectral.nwlP
                                                            - 1]
            soilopt.refl = self._rsoil
            self.soilopt = soilopt
            self._tracker['soil'] = False

        # Run the PROSPECT model
        if self._tracker['leaf']:
            leafopt = PROSPECT_5D(self._leafbio, self.optipar)
            # Update leaf optics refl and trans to include thermal
            # values from model assumptions
            self._rho[self.spectral.IwlP] = leafopt.refl
            self._tau[self.spectral.IwlP] = leafopt.tran
            leafopt.refl = self._rho
            leafopt.tran = self._tau
            self.leafopt = leafopt
            self._tracker['leaf'] = False


        # Run the SAIL model
        rad = SAILH(self.soilopt, self.leafopt, self._canopy, self._angles)
        self.canopyopt = rad

        sensor_wavelengths = self.sensorinfo['wl_smac'].T[0]

        # Interpolate whole wavelength radiances to sensor wavlengths
        rv_so = np.interp(sensor_wavelengths, self.spectral.wlS,
                          rad.rso.T[0])
        rv_do = np.interp(sensor_wavelengths, self.spectral.wlS,
                          rad.rdo.T[0])
        rv_dd = np.interp(sensor_wavelengths, self.spectral.wlS,
                          rad.rdd.T[0])
        rv_sd = np.interp(sensor_wavelengths, self.spectral.wlS,
                          rad.rsd.T[0])

        # Run the SMAC atmosphere model
        if (self._tracker['atm'] or self._tracker['angles'] or
            self._tracker['sensor']):
            atmopt = SMAC(self._angles, self._atm, self.sensorinfo['SMAC_coef'])
            self.atmopt = atmopt
            self._tracker['atm'] = False
            self._tracker['angles'] = False
            self._tracker['sensor'] = False

        # Upscale TOC to TOA
        ta_ss = self.atmopt.Ta_ss
        ta_sd = self.atmopt.Ta_sd
        ta_oo = self.atmopt.Ta_oo
        ta_do = self.atmopt.Ta_do
        ra_dd = self.atmopt.Ra_dd
        ra_so = self.atmopt.Ra_so
        T_g = self.atmopt.Tg

        rtoa0 = ra_so + ta_ss * rv_so * ta_oo
        rtoa1 = (ta_sd * rv_do + ta_ss * rv_sd * ra_dd * rv_do) * ta_oo / \
                (1 - rv_dd * ra_dd)
        rtoa2 = (ta_ss * rv_sd + ta_sd * rv_dd) * ta_do / (1 - rv_dd * ra_dd)
        self.R_TOC = (ta_ss * rv_so + ta_sd * rv_do) / (ta_ss + ta_sd)
        self.R_TOA = T_g * (rtoa0 + rtoa1 + rtoa2)
        self.L_TOA = self._La * self.R_TOA

        bands = self.sensorinfo['band_id_smac']

        results_table = pd.DataFrame(zip(bands, self.L_TOA[0], self.R_TOA[0],
                                         self.R_TOC[0]),
                                     index=sensor_wavelengths,
                                     columns=['Band', 'L_TOA', 'R_TOA', 'R_TOC'])

        return results_table


class SpectralBands:
    """
    Class to hold definitions of spectral band ranges and wavelengths.

    Attributes
    ----------
    wlP : np.array
        Range of wavelengths over which the PROSPECT model operates.
    wlE : np.array
        Range of wavelengths in E-F excitation matrix
    wlF : np.array
        Range of wavelengths for chlorophyll fluorescence in E-F matrix
    wlO : np.array
        Range of wavelengths in the optical part of the spectrum
    wlT : np.array
        Range of wavelengths in the thermal part of the spectrum
    wlS : np.array
        Range of wavelengths for the solar spectrum. wlO and wlT combined.
    wlPAR : np.array
        Range of wavelengths for photosynthetically active radiation
    nwlP : int
        Number of optical bands
    nwlT : int
        Number of thermal bands
    IwlP : range
        Index of optical bands
    IwlT : range
        Index of thermal bands
    """
    def __init__(self):
        self.wlP = np.arange(400, 2401, 1)
        self.wlE = np.arange(400, 751, 1)
        self.WlF = np.arange(640, 851, 1)
        self.wlO = np.arange(400, 2401, 1)
        self.wlT = np.concatenate([np.arange(2500, 15001, 100),
                                   np.arange(16000, 50001, 1000)])
        self.wlS = np.concatenate([self.wlO, self.wlT])
        self.wlPAR = np.arange(400, 701, 1)
        self.nwlP = len(self.wlP)
        self.nwlT = len(self.wlT)
        self.IwlP = np.arange(0, self.nwlP, 1)
        self.IwlT = np.arange(self.nwlP, self.nwlP + self.nwlT, 1)


def calculate_ET_radiance(Ea, DOY, tts):
    """
    Calculate extraterrestrial radiation.

    Parameters
    ----------
    Ea : float
        Solar constant for spectral irradiance
    DOY : int
        Day of year, Julian calendar
    tts : float
        Solar zenith angle in degrees

    Returns
    -------
    np.array
        Solar extraterrestrial spectrum

    NOTE
    ----
    see: https://www.sciencedirect.com/topics/engineering/
            extraterrestrial-radiation
    """
    # NOTE:
    # In the original code the below line contains 'DOY / 265' rather than
    # 'DOY / 365' however this is assumed to be a mistake and is corrected
    # here.
    b = 2 * np.pi * DOY / 365
    correction_factor = 1.00011 + 0.034221 * np.cos(b) + 0.00128 * np.sin(b) \
        + 0.000719 * np.cos(2 * b) + 0.000077 * np.sin(2 * b)
    La = Ea * correction_factor * np.cos(tts * np.pi / 180) / np.pi

    return La


def calculate_spectral_convolution(wl_hi, radiation_spectra, sensorinfo):
    """
    Calculate the spectral convolution for a given spectral response function.

    Parameters
    ----------
    wl_hi : np.array
        Arrays of wavelengths to be convolved
    radiation_spectra : np.array
        irradiance or radiance to be convolved
    sensorinfo : dict
        Contains keys 'wl_srf' -> number of bands contrib * number of bands
        post con
        'p_srf' -> relative contribution of each band

    Returns
    -------
    np.array
        Convolution result
    """
    wl_srf = sensorinfo['wl_srf_smac']
    p_srf = sensorinfo['p_srf_smac']

    def get_closest_index(V, N):
        # Find the indices of the closest entries in N to the values of those
        # in V
        V = np.reshape(V, (V.shape[0] * V.shape[1], 1), order='F')
        A = np.repeat(N, len(V), axis=1)
        closest_index = np.argmin(np.abs(A - V.T), 0)
        return closest_index

    indx = get_closest_index(wl_srf, wl_hi)
    rad = np.reshape(radiation_spectra[indx], (wl_srf.shape[0],
                                               wl_srf.shape[1]),
                     order='F')
    # Sum and normalize as p_srf is not normalized.
    rad_conv = np.sum(rad * p_srf, axis=0) / np.sum(p_srf, axis=0)

    return rad_conv


def load_optical_parameters():
    """Load optical parameters from saved arrays"""
    params_file = Path(__file__).parent / 'model_parameters/optical_params.pkl'

    with open(params_file, 'rb') as f:
        params = pickle.load(f)

    return params


def load_ET_parameters():
    """Load extratrerrestrial parameters from saved arrays"""
    params_file = Path(__file__).parent / 'model_parameters/ET_irradiance.pkl'

    with open(params_file, 'rb') as f:
        params = pickle.load(f)

    return params


def load_sensor_info(sensor):
    """Load RS sensor information from saved arrays"""
    sensor_path = Path(__file__).parent / f'sensor_information/{sensor}.pkl'
    with open(sensor_path, 'rb') as f:
        sensor_info = pickle.load(f)
    return sensor_info


if __name__ == '__main__':
    leafbio = LeafBiology(40, 10, 0.02, 0.01, 0, 10, 1.5)
    soilpar = SoilParameters(0.5, 0, 100, 15)
    canopy = CanopyStructure(0, -0.35, -0.15, 0.05)
    angles = Angles(40, 0, 0)
    atm = AtmosphericProperties(0.3246, 0.3480, 1.4116, 1013.25)
    spart = SPART(soilpar, leafbio, canopy, atm, angles, 'LANDSAT8-OLI',
                  100)
    print(spart.run())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

def parse_requirements(filename):
    """Load requirements from a pip requirements file."""
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

setuptools.setup(
    name="SPART-python",
    version="1.0.9",
    author="G. Worrall",
    author_email="worrall.george@gmail.com",
    description="Python implementation of the SPART optical radiative transfer model.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wirrell/SPART-python",
    install_requires=parse_requirements('requirements.txt'),
    packages=setuptools.find_packages(),
    package_data={
        'SPART': ['model_parameters/*.pkl', 'sensor_information/*.pkl']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS"
    ],
    python_requires='>=3.4',
)



"""
Brightness-Shape-Moisture soil model.

Ported from the original matlat SPART code.

Model as outlined in:
    The SPART model: A soil-plant-atmosphere radiative transfer model
    for satellite measurements in the solar spectrum - Yang et al.
"""
import warnings
import numpy as np
import pandas as pd
from scipy.stats import poisson
from .PROSPECT_5D import calculate_tav


def BSM(soilpar, soilspec):
    """
    Run the BSM soil model

    Parameters
    ----------
    soilpar : SoilParameters
        Object with attributes [B, lat, lon] / dry soil spectra, and SMp, SMC,
        film
    soilspec : dict
        Contains keys ['GSV', 'kw', 'nw'] which key the Global Soil Vectors,
        water absorption constants for the spectrum and water refraction index
        for the spectrum. Loaded in in the main SPART script and passed to this
        function.

    Returns
    -------
    SoilOptics
    """
    # Soil parameters - See object SoilParameters for variable details
    SMp = soilpar.SMp
    SMC = soilpar.SMC  # NOTE: in the original code these are held in
    film = soilpar.film  # a struct named 'emp' and set to constants.

    # Spectral parameters of the soil
    if soilpar.rdry_set:
        rdry = soilpar.rdry
    else:
        GSV = soilspec['GSV']  # Global Soil Vectors spectra
        B = soilpar.B
        lat = soilpar.lat
        lon = soilpar.lon
        f1 = B * np.sin(lat * np.pi / 180)
        f2 = B * np.cos(lat * np.pi / 180) * np.sin(lon * np.pi / 180)
        f3 = B * np.cos(lat * np.pi / 180) * np.cos(lon * np.pi / 180)
        rdry = f1 * GSV[:, [0]] + f2 * GSV[:, [1]] + f3 * GSV[:, [2]]

    kw = soilspec['Kw']  # Water absoprtion specturm
    nw = soilspec['nw']  # Water refraction index spectrum

    rwet = soilwat(rdry, nw, kw, SMp, SMC, film)

    return rwet


def soilwat(rdry, nw, kw, SMp, SMC, deleff):
    """
    Model soil water effects on soil reflectance and return wet reflectance.

    From original matlab code:
        In this model it is assumed that the water film area is built up
        according to a Poisson process.

    See the description in the original model paper in the top of script
    docstring.

    Parameters
    ----------
    rdry : np.array
        Dry soil reflectance
    nw : np.array
        Refraction index of water
    kw : np.array
        Absorption coefficient of water
    SMp : float
        Soil moisture volume [%]
    SMC : float
        Soil moisture carrying capacity
    deleff : float
        Effective optical thickness of single water film, cm

    Returns
    -------
    SoilOptics
        Wet soil reflectance spectra across 400 nm to 2400 nm

    NOTE
    ----
    The original matlab script accepts SMp row vectors for different SM
    percentages. This is not implemented here but may need to be in future
    if there is a significant speed bonus to doing so.
    """
    k = [0, 1, 2, 3, 4, 5, 6]
    nk = len(k)
    mu = (SMp - 5) / SMC
    if mu <= 0:  # below 5 % SM -> model assumes no effect
        rwet = rdry
    else:
        # From original matlab: Lekner & Dorf (1988)
        #
        # Uses t_av calculation from PROSPECT-5D model. If you want BSM
        # script to operate independently you will need to paste that
        # function in here.
        rbac = 1 - (1 - rdry) * (rdry * calculate_tav(90, 2 / nw) /
                                 calculate_tav(90, 2) + 1 - rdry)

        # total reflectance at bottom of water film surface
        p = 1 - calculate_tav(90, nw) / nw ** 2

        # reflectance of water film top surface, use 40 degrees incidence angle
        # like in PROSPECT (note from original matlab script)
        Rw = 1 - calculate_tav(40, nw)

        fmul = poisson.pmf(k, mu)
        tw = np.exp(-2 * kw * deleff * k)
        Rwet_k = Rw + (1 - Rw) * (1 - p) * tw * rbac / (1 - p * tw * rbac)
        rwet = (rdry * fmul[0]) + Rwet_k[:, 1:nk].dot(fmul[1:nk])[:,
                                                                  np.newaxis]

    soilopt = SoilOptics(rwet, rdry)

    return soilopt


class SoilOptics:
    """
    Class to hold soil optical reflectance spectra.

    Parameters
    ----------
    refl : np.array
        Soil reflectance spectra (with SM taken into account)
    refl_dry : np.array
        Dry soil reflectance spectra

    Attributes
    ----------
    refl : np.array
        Soil reflectance spectra (with SM taken into account)
    refl_dry : np.array
        Dry soil reflectance spectra
    """
    def __init__(self, refl, refl_dry):
        self.refl = refl
        self.refl_dry = refl_dry


class SoilParametersFromFile:
    """
    Class to load and hold soil reflectance spectrum from the JPL soil
    reflectance data available at https://speclib.jpl.nasa.gov/

    Parameters
    ----------
    soil_file : str or np.array
        path to JPL soil reflectance spectra file or array containing rdry
    SMp : float
        Soil moisture percentage [%]
    SMC : float, optional
        Soil moisture carrying capacity of the soil
    film : float, optional
        Single water film optical thickness, cm

    Attributes
    ----------
    rdry_set : bool
        True. Declares that the object contains a dry soil reflectance spectra
    rdry : np.array
        Array containing soil reflectance spectrum extracted from file and
        interpolated to 1 nm intervals between 400 nm and 2400 nm
    """
    def __init__(self, soil_file, SMp, SMC=None, film=None):
        if isinstance(SMC, type(None)):
            warnings.warn("BSM soil model: SMC not supplied, set to default of 25 %")
            self.SMC = 25
        else:
            self.SMC = SMC
        if isinstance(film, type(None)):
            warnings.warn("BSM soil model: water film optical thickness not supplied,")
            warnings.warn("\t set to default of 0.0150 cm")
            self.film = 0.0150
        else:
            self.film = film

        # Allow the rdry spectra to be passed directly to the object.
        if isinstance(soil_file, np.ndarray):
            self.rdry = soil_file
        else:
            self.rdry = self._load_jpl_soil_refl(soil_file)
        self.SMp = SMp
        self.rdry_set = True

    def _load_jpl_soil_refl(self, file_path):
        """Load and format the JPL supplied soil reflectance file."""
        # Load in table
        soil_refl = pd.read_csv(file_path, sep='\t', skiprows=21,
                                index_col=0, header=None)
        # turn micrometers index to nanometers
        soil_refl.index = soil_refl.index * 1000
        # convert percentage reflectance to fraction
        if (soil_refl.loc[:, 1] > 1).any():
            soil_refl = soil_refl / 100

        # Get only 400 nm to 2400 nm and order so 400 nm comes first
        soil_refl = soil_refl[2401:400][::-1]

        # Interpolate so reflectances on whole nm values
        wls = np.arange(400, 2401, 1)
        for wl in wls:
            if not wl in soil_refl.index:
                soil_refl.loc[wl, 1] = np.nan
        soil_refl = soil_refl.sort_index()

        soil_refl = soil_refl.interpolate('linear')
        soil_refl = soil_refl.loc[wls].to_numpy()

        return soil_refl


class SoilParameters:
    """
    Class to hold the soil characteristics for BSM.

    Parameters
    ----------
    B : float
        Soil brightness as defined in the paper.
    lat : float
        Soil spectral coordinate, latitiude, realistic range 80 - 120 deg
        for soil behavior (see paper, phi)
    lon : float
        Soil spectral coordinate, longitude, realistic range -30 - 30 deg
        for soil behaviour (see paper, lambda)
    SMp : float
        Soil moisture percentage [%]
    SMC : float, optional
        Soil moisture carrying capacity of the soil
    film : float, optional
        Single water film optical thickness, cm

    Attributes
    ----------
    B : float
        Soil brightness as defined in the paper.
    lat : float
        Soil spectral coordinate, latitiude
    lon : float
        Soil spectral coordinate, longitude
    SMp : float
        Soil moisture percentage [%]
    SMC : float
        Soil moisture carrying capacity of the soil
    film : float
        Single water film optical thickness, cm
    rdry_set : bool
        False. Declares that the object doesnt' contain a dry soil reflectance
        spectra
    """
    def __init__(self, B, lat, lon, SMp, SMC=None, film=None):
        self.B = B
        self.lat = lat
        self.lon = lon
        self.SMp = SMp
        if isinstance(SMC, type(None)):
            warnings.warn("BSM soil model: SMC not supplied,"
                          " set to default of 25 %")
            self.SMC = 25
        else:
            self.SMC = SMC
        if isinstance(film, type(None)):
            warnings.warn("BSM soil model: water film optical thickness"
                          " not supplied, set to default of 0.0150 cm")
            self.film = 0.0150
        else:
            self.film = film
        self.rdry_set = False


if __name__ == '__main__':
    # Test cases, compare to original matlab outputs
    from SPART import load_optical_parameters
    soilpar = SoilParameters(0.5, 0, 100, 15)
    soilopt = BSM(soilpar, load_optical_parameters())
    print(soilopt.refl)
